{ ----------------------------------------------------------------------------}
{ A Gradient Fill component for Delphi.                                       }
{ Copyright 1995, Curtis White.  All Rights Reserved.                         }
{ This component can be freely used and distributed in commercial and private }
{ environments, provied this notice is not modified in any way.               }
{ ----------------------------------------------------------------------------}
{ Feel free to contact me if you have any questions, comments or suggestions  }
{ at cwhite@teleport.com                                                      }
{ ----------------------------------------------------------------------------}
{ Date last modified:  12/28/96                                               }
{ ----------------------------------------------------------------------------}
{ ----------------------------------------------------------------------------}
{ TGradientFill v1.35                                                         }
{ ----------------------------------------------------------------------------}
{ Description:                                                                }
{   A graphic control that displays a gradient beginning with a chosen color  }
{   and ending with another chosen color.                                     }
{ Features:                                                                   }
{   The begin and end colors can be any colors.                               }
{   The fill direction can be set to Top-To-Bottom, Bottom-To-Top,            }
{     Right-To-Left, or Left-To-Right.                                        }
{   The number of colors, between 1 and 255 can be set for the fill.          }
{ ----------------------------------------------------------------------------}
{ ----------------------------------------------------------------------------}
{ Revision History:                                                           }
{ 1.00:  Initial release                                                      }
{ 1.10:  Added routines to use a realized palette for systems less than true  }
{        color.  This gets rid of Windows 16 color dithering on systems that  }
{        can display only 256 or less colors.                                 }
{ 1.20:  Added two new properties to allow the Gradient Fill component to be  }
{        used as a background in MDI forms.  The fill paints to a bitmap,     }
{        which is a published property, then copies the bitmap to the canvas. }
{        This allows the bitmap to be copied to the background of an MDI form.}
{        A palette handle property has also been added to allow the palette   }
{        to be realized by the forms canvas when the Realize property is true.}
{ 1.30:  Fixed resource leak.  Added a flag to allow properties to be set     }
{        without repainting each time a property is set.  Then this flag can  }
{        be set and the gradient fill will repaint itself with all the new    }
{        changes in place.  Also added a new method to allow it to realise    }
{        it's palette as a background palette.  The method is designed to be  }
{        called by the owning app if it gets a WM_PALETTECHANGED message to   }
{        try and keep the colors as close as possible on a 256 system.  This  }
{        routine was contributed by Colin Messitt (74774.1347@compuserve.com) }
{ 1.35:  Changed the About box and made it compatible with Delphi 2.0         }
{        (32-bit version). Removed the component editor for the about box.    }
{ ----------------------------------------------------------------------------}

unit Gfvcl;

interface

{$D-,Y-,L-,U-,B-}

uses
  WinTypes, WinProcs, Classes, Graphics, Controls, Menus, Forms, DsgnIntf, GFAbout;
{  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls, Menus,
  Forms, Dialogs, ExtCtrls, DsgnIntf, GFAbout;}

type
  { Direction of fill }
  TFillDirection = (fdTopToBottom, fdBottomToTop, fdLeftToRight, fdRightToLeft);
  { Range of valid colors }
  TNumberOfColors = 1..256;

  TGradientFill = class(TGraphicControl)
  private
    { Variables for properties }
    FDirection: TFillDirection;
    FBeginColor: TColor;
    FEndColor: TColor;
    FAutoSize: Boolean;
    FRealize: Boolean;
    FNumberOfColors: TNumberOfColors;
    FNumColors: TNumberOfColors;
    FPicture: TBitmap;
    FHoldRedraw: Boolean;
    FGFAbout: TGFAboutBox;
    FLogPal: PLogPalette;  {Pointer to a TLogPalette}
    FPalSize: LongInt;     {Used to calculate memory to allocate for TLogPalette}
    FPalette : HPalette;   {Used to store Palette created with LogPal}

    { Procedures for setting property values }
    procedure SetFillDirection(Value: TFillDirection);
    procedure SetAutoSize(Value: Boolean);
    procedure SetRealize(Value: Boolean);
    procedure SetBeginColor(Value: TColor);
    procedure SetEndColor(Value: TColor);
    procedure SetNumberOfColors(Value: TNumberOfColors);
    procedure SetHoldRedraw(Value: Boolean);

    { Fill procedure }
    procedure GradientFill;
  protected
    procedure Paint; override;
    procedure MakePalette(BeginRed, BeginGreen, BeginBlue: Byte;
      RedDifference, GreenDifference, BlueDifference: Integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ReRealize(OrigWnd: hWnd);
    property Picture: TBitmap read FPicture;
    property PaletteHandle: HPalette read FPalette;
  published
    { AboutBox }
    property About: TGFAboutBox read FGFAbout write FGFAbout;
    { Repaint when autosized }
    property AutoSize: Boolean read FAutoSize write SetAutoSize default False;
    { Starting color of fill }
    property BeginColor: TColor read FBeginColor write SetBeginColor
        default clBlue;
    { Ending color of fill }
    property EndColor: TColor read FEndColor write SetEndColor default clBlack;
    { Direction of fill }
    property FillDirection: TFillDirection read FDirection
        write SetFillDirection default fdTopToBottom;
    { Property to determine if redraw should be done or not.  This allows
      multiple changes to be made without redrawing the Gradient Fill }
    property HoldRedraw: Boolean read FHoldRedraw write SetHoldRedraw
        default False;
    { Number of colors to use in the fill (1 - 256) - default is 16.  If 1 }
    { then it uses the Begin Color.                                        }
    property NumberOfColors: TNumberOfColors read FNumColors
        write SetNumberOfColors default 16;
    { Use a Realized Palette if True }
    property Realize: Boolean read FRealize write SetRealize default True;
    { Enable standard properties }
    property Align;
    property DragCursor;
    property DragMode;
    property Enabled;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

{--------------------------------------------------------------------}

  { Component AboutBox }
{  TGFCompAbout = class(TComponentEditor)
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
    procedure ShowAbout;
  end;
}
{--------------------------------------------------------------------}

  { Property AboutBox }
  TGFPropAbout = class(TClassProperty)
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    procedure ShowAbout;
  end;

procedure Register;

implementation

var
  GFAboutBox: TGFAboutBox;

{ TGradientFill }

{ Override the constructor to initialize variables }
constructor TGradientFill.Create(AOwner: TComponent);
begin
  { Inherit original constructor }
  inherited Create(AOwner);
  { Add new initializations }
  Height := 105;
  Width := 105;
  FBeginColor := clBlue;
  FEndColor := clBlack;
  FDirection := fdTopToBottom;
  FNumberOfColors := 16;
  FNumColors := FNumberOfColors;
  FRealize := True;
  FPicture := TBitmap.Create;
end;

{ Override the destructor to clean up }
destructor TGradientFill.Destroy;
begin
  { Add clean up }
  { if (FPalette > 0) and
     (not (csDesigning in ComponentState)) then  }
  if (FPalette > 0) then
    DeleteObject(FPalette);
  FPicture.Free;
  { Inherit original destructor }
  inherited Destroy;
end;

{ Set begin color when property is changed }
procedure TGradientFill.SetBeginColor(Value: TColor);
begin
  if Value <> FBeginColor then
  begin
    FBeginColor := Value;
    if not FHoldRedraw then
    begin
      GradientFill;
      Refresh;
    end;
  end;
end;

{ Set end color when property is changed }
procedure TGradientFill.SetEndColor(Value: TColor);
begin
  if Value <> FEndColor then
  begin
    FEndColor := Value;
    if not FHoldRedraw then
    begin
      GradientFill;
      Refresh;
    end;
  end;
end;

{ Repaint the screen upon a resize }
procedure TGradientFill.SetAutoSize(Value: Boolean);
begin
  if Value <> FAutoSize then
  begin
    FAutoSize := Value;
    Repaint;
  end;
end;

{ Use a Realized Palette if True }
procedure TGradientFill.SetRealize(Value: Boolean);
begin
  if Value <> FRealize then
  begin
    FRealize := Value;
    if not FHoldRedraw then
    begin
      GradientFill;
      Refresh;
    end;
  end;
end;

{ Set the number of colors to be used in the fill }
procedure TGradientFill.SetNumberOfColors(Value: TNumberOfColors);
begin
  if Value <> FNumColors then
  begin
    FNumColors := Value;
    if FRealize and (Value = 256) then
      { Subtract one since 256 leaves an off color bar in the middle }
      FNumberOfColors := Value - 1
    else
      FNumberOfColors := Value;

    if not FHoldRedraw then
    begin
      GradientFill;
      Refresh;
    end;
  end;
end;

{ Set the fill direction }
procedure TGradientFill.SetFillDirection(Value: TFillDirection);
begin
  if Value <> FDirection then
  begin
    FDirection := Value;
    if not FHoldRedraw then
    begin
      GradientFill;
      Refresh;
    end;
  end;
end;

{ Set the HoldRedraw property }
procedure TGradientFill.SetHoldRedraw(Value: Boolean);
begin
  if Value <> FHoldRedraw then
  begin
    { Set the value }
    FHoldRedraw := Value;
    { If the hold is turned off, then redraw the Gradient }
    if not FHoldRedraw then
    begin
      GradientFill;
      Refresh;
    end;
  end;
end;

{ Perform the fill when paint is called }
procedure TGradientFill.Paint;
begin
  if not FHoldRedraw then
  begin
    GradientFill;
    inherited Paint;
  end;
end;

{ Build the Palette }
procedure TGradientFill.MakePalette(BeginRed, BeginGreen, BeginBlue: Byte;
   RedDifference, GreenDifference, BlueDifference: Integer);
var
  I : Integer;  { For Loop that changes color}
  R : Byte;     { Color band Red value }
  G : Byte;     { Color band Green value }
  B : Byte;     { Color band Blue value }
begin
  { Initialize Colors }
  R := BeginRed;
  G := BeginGreen;
  B := BeginBlue;
  { Delete the palette object if one is already active }
  if FPalette > 0 then
    DeleteObject(FPalette);

  { 2 *SizeOf(Word) to get space for palVersion and palNumEntries }
  FPalSize := (2 * SizeOf(Word)) + ((FNumberOfColors + 1) * SizeOf(TPaletteEntry));
  { Alocate memory for the palette }
  GetMem(FLogPal, FPalSize);
  try
    { Set the palette version }
    FLogPal^.palVersion:=$0300;
    { Set the number of palette entries }
    FLogPal^.palNumEntries := FNumberOfColors + 1;

    { Create each TLogPaletteEntry }
    for I := 0 to FNumberOfColors do
    begin
      { Set RED amount }
      FLogPal^.palPalEntry[I].peRed := R;
      { Set GREEN amount }
      FLogPal^.palPalEntry[I].peGreen := G;
      { Set Blue amount }
      FLogPal^.palPalEntry[I].peBlue := B;
      { Set Flag }
      FLogPal^.palPalEntry[I].peFlags := 0;

      { Calculate the next color }
      if FNumberOfColors > 1 then
      begin
        R := BeginRed + MulDiv (I, RedDifference, FNumberOfColors - 1);
        G := BeginGreen + MulDiv (I, GreenDifference, FNumberOfColors - 1);
        B := BeginBlue + MulDiv (I, BlueDifference, FNumberOfColors - 1);
      end
      else
      { Set to the Begin Color if set to only one color }
      begin
        R := BeginRed;
        G := BeginGreen;
        B := BeginBlue;
      end;
    end;

    { Create the palette }
    FPalette := CreatePalette(FLogPal^);           {Call API to make palette from LogPal}
  finally
    { Free the memory }
    FreeMem(FLogPal, FPalSize);                     {Destroy LogPal in memory}
  end;
end;

{ Gradient fill procedure - the actual routine }
procedure TGradientFill.GradientFill;
var
  { Set up working variables }
  BeginRGBValue  : array[0..2] of Byte;    { Begin RGB values }
  RGBDifference  : array[0..2] of integer; { Difference between begin and end }
                                           { RGB values                       }
  ColorBand : TRect;    { Color band rectangular coordinates }
  I         : Integer;  { Color band index }
  J         : Integer;  { Color band index for Reverse }
  R         : Byte;     { Color band Red value }
  G         : Byte;     { Color band Green value }
  B         : Byte;     { Color band Blue value }
begin
  { Create the bitmap width and height }
  FPicture.Width := Width;
  FPicture.Height := Height;

  { Use working bitmap to draw the gradient }
  with FPicture do
  begin
    { Extract the begin RGB values }
    { Set the Begin Red, Green and Blue colors }
    BeginRGBValue[0] := GetRValue (ColorToRGB (FBeginColor));
    BeginRGBValue[1] := GetGValue (ColorToRGB (FBeginColor));
    BeginRGBValue[2] := GetBValue (ColorToRGB (FBeginColor));
    { Calculate the difference between begin and end RGB values }
    RGBDifference[0] := GetRValue (ColorToRGB (FEndColor)) -
                        BeginRGBValue[0];
    RGBDifference[1] := GetGValue (ColorToRGB (FEndColor)) -
                        BeginRGBValue[1];
    RGBDifference[2] := GetBValue (ColorToRGB (FEndColor)) -
                        BeginRGBValue[2];

    { Set the pen style and mode }
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Mode := pmCopy;

    if Realize then
    begin
      { Build the Palette }
      MakePalette(BeginRGBValue[0], BeginRGBValue[1], BeginRGBValue[2],
         RGBDifference[0], RGBDifference[1], RGBDifference[2]);
      { Set Canvas's palette to the new one }
      SelectPalette(Canvas.Handle, FPalette, False);
      { Make Canvas recognize the palette }
      RealizePalette(Canvas.Handle);
    end;

    case FDirection of
      { Calculate the color band's top and bottom coordinates }
      { for TopToBottom and BottomToTop fills }
      fdTopToBottom, fdBottomToTop:
        begin
          ColorBand.Left := 0;
          ColorBand.Right := Width;
        end;

      { Calculate the color band's left and right coordinates }
      { for LeftToRight and RightToLeft fills }
      fdLeftToRight, fdRightToLeft:
        begin
          ColorBand.Top := 0;
          ColorBand.Bottom := Height;
        end;
    end;

    case FDirection of
      fdTopToBottom, fdLeftToRight:
        begin
          { Perform the fill }
          for I := 0 to FNumberOfColors do
          begin
            case FDirection of
              { Calculate the color band's top and bottom coordinates }
              fdTopToBottom:
                begin
                  ColorBand.Top    := MulDiv (I, Height, FNumberOfColors);
                  ColorBand.Bottom := MulDiv (I + 1, Height, FNumberOfColors);
                end;

              { Calculate the color band's left and right coordinates }
              fdLeftToRight:
                begin
                  ColorBand.Left  := MulDiv (I, Width, FNumberOfColors);
                  ColorBand.Right := MulDiv (I + 1, Width, FNumberOfColors);
                end;
            end;

            { Calculate the color band's color }
            if FNumberOfColors > 1 then
            begin                                                          { Took away -1 on FNumberOfColors }
              R := BeginRGBValue[0] + MulDiv (I, RGBDifference[0], FNumberOfColors);
              G := BeginRGBValue[1] + MulDiv (I, RGBDifference[1], FNumberOfColors);
              B := BeginRGBValue[2] + MulDiv (I, RGBDifference[2], FNumberOfColors);
            end
            else
            { Set to the Begin Color if set to only one color }
            begin
              R := BeginRGBValue[0];
              G := BeginRGBValue[1];
              B := BeginRGBValue[2];
            end;

            if Realize then
              { Set the color from the palette }
              Canvas.Brush.Color := PaletteRGB(R, G, B)
            else
              { Select the calculated color into the brush  }
              Canvas.Brush.Color := RGB (R, G, B);

            { Paint the color band }
            Canvas.FillRect (ColorBand);
          end;
        end;
      fdBottomToTop, fdRightToLeft:
        begin
          { Initialize up counter }
          J := 0;
          { Perform the fill }
          for I := FNumberOfColors downto 0 do
          begin
            case FDirection of
              { Calculate the color band's top and bottom coordinates }
              fdBottomToTop:
                begin
                  ColorBand.Top    := MulDiv (I - 1, Height, FNumberOfColors);
                  ColorBand.Bottom := MulDiv (I, Height, FNumberOfColors);
                end;

              { Calculate the color band's left and right coordinates }
              fdRightToLeft:
                begin
                  ColorBand.Left  := MulDiv (I - 1, Width, FNumberOfColors);
                  ColorBand.Right := MulDiv (I, Width, FNumberOfColors);
                end;
            end;

            { Calculate the color band's color }
            if FNumberOfColors > 1 then
            begin                                                          { Took away -1 on FNumberOfColors }
              R := BeginRGBValue[0] + MulDiv (J, RGBDifference[0], FNumberOfColors);
              G := BeginRGBValue[1] + MulDiv (J, RGBDifference[1], FNumberOfColors);
              B := BeginRGBValue[2] + MulDiv (J, RGBDifference[2], FNumberOfColors);
            end
            else
            { Set to the Begin Color if set to only one color }
            begin
              R := BeginRGBValue[0];
              G := BeginRGBValue[1];
              B := BeginRGBValue[2];
            end;

            if Realize then
              { Set the color from the palette }
              Canvas.Brush.Color := PaletteRGB(R, G, B)
            else
              { Select the calculated color into the brush  }
              Canvas.Brush.Color := RGB (R, G, B);

            { Increment J for color calculations }
            Inc(J);

            { Paint the color band }
            Canvas.FillRect (ColorBand);
          end;
        end;
    end;
  end;

  { Set Canvas's palette to the new one }
  SelectPalette(Canvas.Handle, FPalette, False);
  { Make Canvas recognize the palette }
  RealizePalette(Canvas.Handle);
  { Copy the bitmap to the main canvas }
  Canvas.Draw(0, 0, FPicture);
end;

{ Method to Re-Realize the Palette }
procedure TGradientFill.ReRealize(OrigWnd: hWnd);
begin
  { Only do if the Windows handle is different
    than the Gradient Fills handle }
  if OrigWnd <> Canvas.Handle then
  begin
    { Set Canvas's palette to the new one }
    SelectPalette(Canvas.Handle, FPalette, False);
    { Make Canvas recognize the palette }
    RealizePalette(Canvas.Handle);
  end;
end;

{---------------------------------------------------------------------}

{ Component AboutBox }
{procedure TGFCompAbout.ShowAbout;
begin
  GFAboutBox := TGFAboutBox.Create(Application);
  GFAboutBox.ShowModal;
  GFAboutBox.Free;
end;

function TGFCompAbout.GetVerbCount: Integer;
begin
  Result := 1;
end;

function TGFCompAbout.GetVerb(Index: Integer): string;
begin
  Result := 'About';
end;

procedure TGFCompAbout.ExecuteVerb(Index: Integer);
begin
  ShowAbout;
end;
}
{---------------------------------------------------------------------}

{ Property AboutBox }
procedure TGFPropAbout.ShowAbout;
begin
  GFAboutBox := TGFAboutBox.Create(Application);
  GFAboutBox.ShowModal;
  GFAboutBox.Free;
end;

procedure TGFPropAbout.Edit;
begin
  ShowAbout;
end;

function TGFPropAbout.GetAttributes: TPropertyAttributes;
begin
  Result := [paMultiSelect, paDialog, paReadOnly];
end;

{ Register the component }
procedure Register;
begin
  RegisterComponents('TechnoSoft', [TGradientFill]);
{  RegisterComponentEditor(TGradientFill, TGFCompAbout);}
  RegisterPropertyEditor(TypeInfo(TGFAboutBox), TGradientFill, 'About', TGFPropAbout);
end;

end.
